# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/10/02
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from docutils import utils, nodes, core
from docutils.parsers.rst.directives.html import MetaBody
from pyramid_controllers.util import getVersion
from .i18n import _

#------------------------------------------------------------------------------
def rtitle(text, *args, **kw):
  return nodes.title('', '', text, *args, **kw)
def rpara(*args, **kw):
  return nodes.paragraph('', '', *args, **kw)
def rtext(text, *args, **kw):
  return nodes.Text(text, *args, **kw)
def rsect(title, *args, **kw):
  return nodes.section('', rtitle(rtext(title)), *args, **kw)
def rcont(*args, **kw):
  return nodes.container('', *args, **kw)
Meta = MetaBody('').meta
def rmeta(*args, **kw):
  return Meta('', *args, **kw)

#------------------------------------------------------------------------------
def rst2fragments(text):
  if not text:
    return []
  # todo: this generates error output to STDERR, eg:
  #   <string>:12: (WARNING/2) Inline strong start-string without end-string.
  # it would be *great* if '<string>' were replaced with the actual error
  # source, i.e. controller/method/docstring/etc.

  # TODO: optional failure behaviour: convert the entire
  #       text to a paragraph via:
  #         return [rpara(rtext(text))]
  #       ==> requires error detection.

  return list(core.publish_doctree(text))

# TODO: perhaps all these functions would be much simpler if rstMax was
#       checked once, at exit, and if false, all classes and ids were
#       removed?...

#------------------------------------------------------------------------------
def render(data):
  doc = utils.new_document('<pyramid_describe.document>')
  # todo:
  # doc.settings.text_width = data.options.width

  title = data.options.title or _('Contents of "{}"', data.root)
  if data.options.rstMax:
    doc['title'] = title

  # todo: test this bullet_list stuff (i.e. set showInfo to false)

  endpoints = rcont(rtitle(rtext(title)))
  epcont = endpoints
  if data.options.rstMax:
    endpoints['ids'] = ['section-endpoints']
    endpoints['classes'] = ['endpoints']
  if not data.options.showInfo:
    epcont = nodes.bullet_list('', bullet='*')
    endpoints.append(epcont)
  for endpoint in data.endpoints:
    epcont.append(render_entry(data, endpoint))
  doc.append(endpoints)

  if data.options.showLegend:
    legend = rcont(rtitle(rtext(_('Legend'))))
    if data.options.rstMax:
      legend['ids'] = ['section-legend']
      legend['classes'] = ['legend']
    for item, desc in data.legend:
      section = rsect(item, rpara(rtext(desc)))
      if data.options.rstMax:
        section['ids'] = ['legend-item-' + data.options.idEncoder(item)]
        section['classes'] = ['legend-item']
      legend.append(section)
    doc.append(legend)

  if data.options.showMeta:
    meta = rcont()
    meta.append(rmeta(name='title', content=title))
    if data.options.showGenerator:
      gen = 'pyramid-describe'
      if data.options.showGenVersion:
        gen += '/' + getVersion('pyramid_describe')
      gen += ' [format={}]'.format(data.options.formatstack[-1])
      meta.append(rmeta(name='generator', content=gen))
    if data.options.showLocation and data.options.context.request.url:
      meta.append(rmeta(name='location', content=data.options.context.request.url))
    if data.options.rstMax and data.options.rstPdfkit:
      meta.append(rmeta(name='pdfkit-page-size', content=data.options.pageSize))
      meta.append(rmeta(name='pdfkit-orientation', content=data.options.pageOrientation))
      if not data.options.showOutline:
        meta.append(rmeta(name='pdfkit-no-outline', content=''))
      meta.append(rmeta(name='pdfkit-margin-top', content=data.options.pageMarginTop))
      meta.append(rmeta(name='pdfkit-margin-right', content=data.options.pageMarginRight))
      meta.append(rmeta(name='pdfkit-margin-bottom', content=data.options.pageMarginBottom))
      meta.append(rmeta(name='pdfkit-margin-left', content=data.options.pageMarginLeft))
      if data.options.pageGrayscale:
        meta.append(rmeta(name='pdfkit-grayscale', content=''))
      # todo: other pdfkit meta tags:
      #  meta.append(rmeta(name='pdfkit-print-media-type', content=''))
      #  meta.append(rmeta(name='pdfkit-disable-plugins', content=''))
      #  meta.append(rmeta(name='pdfkit-zoom', content='1.0'))
      #  meta.append(rmeta(name='pdfkit-javascript-delay', content='1000'))
      #  meta.append(rmeta(name='pdfkit-disable-javascript', content=''))
    doc.append(meta)

  return doc

#------------------------------------------------------------------------------
def render_entry(data, entry):
  if not data.options.showInfo:
    return nodes.list_item('', rpara(rtext(entry.dpath)))
  section = rsect(entry.dpath)
  if data.options.rstMax:
    section['ids'] = [entry.id]
    section['classes'] = ['endpoint']
    if entry.classes:
      section['classes'] += entry.classes
  section.extend(render_entry_body(data, entry))
  return section

#------------------------------------------------------------------------------
def render_entry_body(data, entry):
  ret = []
  if data.options.showImpl and entry.ipath:
    impl = _(
      'Handler: {}{} [{}]',
      entry.ipath,'()' if entry.itype == 'instance' else '', entry.itype)
    impl = rpara(rtext(impl))
    if data.options.rstMax:
      impl = rcont(impl)
      impl['ids'] = ['handler-' + entry.id]
      impl['classes'] = ['handler']
    ret.append(impl)
  ret.extend(rst2fragments(entry.doc))
  ret.extend(render_entry_methods(data, entry))
  ret.extend(render_entry_params(data, entry))
  ret.extend(render_entry_returns(data, entry))
  ret.extend(render_entry_raises(data, entry))
  return ret

#------------------------------------------------------------------------------
def render_entry_methods(data, entry):
  if not data.options.showRest or not entry.methods:
    return []
  section = rsect(_('Methods'))
  if data.options.rstMax:
    section['ids'] = ['methods-' + entry.id]
    section['classes'] = ['methods']
  for meth in entry.methods:
    msect = rsect(meth.method or meth.name)
    if data.options.rstMax:
      msect['ids'] = [meth.id]
      msect['classes'] = ['method']
      if meth.classes:
        msect['classes'] += meth.classes
    msect.extend(render_entry_body(data, meth))
    section.append(msect)
  return [section]

#------------------------------------------------------------------------------
def render_entry_params(data, entry):
  if not entry.params:
    return []
  section = rsect(_('Parameters'))
  if data.options.rstMax:
    section['ids'] = ['params-' + entry.id]
    section['classes'] = ['params']
  for node in entry.params:
    rnode = rsect(node.name)
    if data.options.rstMax:
      rnode['ids'] = ['param-' + entry.id + '-' + data.options.idEncoder(node.name)]
      rnode['classes'] = ['param']
      if node.classes:
        rnode['classes'] += node.classes
    spec = []
    if node.type:
      spec.append(node.type)
    if node.optional:
      spec.append(_('optional'))
    if node.default:
      spec.append(_('default {}', node.default))
    spec = _(', ').join(spec)
    if spec:
      spec = rpara(rtext(spec))
      if data.options.rstMax:
        spec = rcont(spec)
        spec['classes'] = ['spec']
      rnode.append(spec)
    rnode.extend(rst2fragments(node.doc))
    section.append(rnode)
  return [section]

#------------------------------------------------------------------------------
def render_entry_returns(data, entry):
  if not entry.returns:
    return []
  section = rsect(_('Returns'))
  if data.options.rstMax:
    section['ids'] = ['returns-' + entry.id]
    section['classes'] = ['returns']
  for node in entry.returns:
    rnode = rsect(node.type)
    if data.options.rstMax:
      rnode['ids'] = ['return-' + entry.id + '-' + data.options.idEncoder(node.type)]
      rnode['classes'] = ['return']
      if node.classes:
        rnode['classes'] += node.classes
    rnode.extend(rst2fragments(node.doc))
    section.append(rnode)
  return [section]

#------------------------------------------------------------------------------
def render_entry_raises(data, entry):
  if not entry.raises:
    return []
  section = rsect(_('Raises'))
  if data.options.rstMax:
    section['ids'] = ['raises-' + entry.id]
    section['classes'] = ['raises']
  for node in entry.raises:
    rnode = rsect(node.type)
    if data.options.rstMax:
      rnode['ids'] = ['raise-' + entry.id + '-' + data.options.idEncoder(node.type)]
      rnode['classes'] = ['raise']
      if node.classes:
        rnode['classes'] += node.classes
    rnode.extend(rst2fragments(node.doc))
    section.append(rnode)
  return [section]

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/10/02
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import re, textwrap
from docutils import core, utils, nodes, writers
import docutils.writers
from docutils.utils.urischemes import schemes

#------------------------------------------------------------------------------
# TODO: these regex's... their necessity needs to be re-evaluated...

enumlistfmt_re = re.compile(
  r'^\(?([0-9]+|[a-z]|[ivxlcdm]+)[.)]\s', flags=re.IGNORECASE)

plaintexturi_re = re.compile(
  r'^('
  + '|'.join([re.escape(s) for s in schemes.keys()])
  + '):'
  + r'([A-Za-z0-9_.~!*\'();:@&=+$,/?#[\]-]*)'
  + '$')

#------------------------------------------------------------------------------
# TODO: all calls to `rstEscape` need to be revisited...
def rstEscape(text, context=None):
  if context in ('`',):
    if re.match('^[a-zA-Z0-9]+$', text):
      return text
    text = text.replace('\\', '\\\\').replace(context, '\\' + context)
    return context + text + context
  if context in (':',):
    if context not in text and '\\' not in text:
      return text
    return text.replace('\\', '\\\\').replace(context, '\\' + context)

  # todo: it is a bit unclear what *should* be done here... but what
  #       is happening now, is to backslash-escape any text that could
  #       be interpreted as a title under/overline... which is primarily
  #       being used by the `visit_bullet_list` method, to avoid creating:
  #         ``* /``
  #       which generates the warning:
  #         Unexpected possible title overline or transition.
  #         Treating it as ordinary text because it's so short.

  if context != 'para' and len(text) > 0 \
      and text == text[0] * len(text) \
      and not re.match('a-zA-Z0-9', text[0]):
    text = ( '\\' + text[0] ) * len(text)
  else:
    if enumlistfmt_re.match(text):
      text = '\\' + text
  return text

#------------------------------------------------------------------------------
def rstTicks(text):
  return rstEscape(text, context='`')

#------------------------------------------------------------------------------
DEFAULT_SECTION_CHARS = '''=-`:'"~^_*+#<>'''
DEFAULT_TEXT_WIDTH    = 79
DEFAULT_INDENT        = ' '*4

#------------------------------------------------------------------------------
class Writer(writers.Writer):

  supported = ('reStructuredText', 'text', 'txt', 'rst')
  'Formats this writer supports.'

  settings_spec = (
    '"Docutils Canonical reStructuredText" Writer Options',
    None,
    (('Specify the order of over/underline characters.',
      ['--section-chars'],
      {'dest': 'section_chars', 'action': 'store',
       'default': DEFAULT_SECTION_CHARS}),
     ('Set the default indentation text.',
      ['--indent'],
      {'dest': 'indent', 'action': 'store',
       'default': DEFAULT_INDENT}),
     ('Set the text wrapping width.',
      ['--text-width'],
      {'dest': 'text_width', 'action': 'store', 'type': 'int',
       'default': DEFAULT_TEXT_WIDTH}),
     ('Output an explicit `title` directive (even if inferrable).',
      ['--explicit-title'],
      {'dest': 'explicit_title', 'action': 'store_true',
       'default': False}),
     ),)

  config_section = 'rst writer'
  config_section_dependencies = ('writers',)

  output = None
  'Final translated form of `document`.'

  def __init__(self):
    writers.Writer.__init__(self)
    self.translator_class = RstTranslator

  def translate(self):
    self.visitor = visitor = self.translator_class(self.document)
    self.document.walkabout(visitor)
    # todo: translate EOL's here?...
    self.output = visitor.output.data()

#------------------------------------------------------------------------------
def collapseLines(value, step):
  if not step:
    return value
  if step != '\n':
    return value + [step]
  if not value:
    return value
  if len(value) > 2 and value[-1] == '\n' and value[-2] == '\n':
    return value
  return value + [step]

#------------------------------------------------------------------------------
newline_re = re.compile('\n([^\n])')
class Output:
  def __init__(self):
    self.lines = []
  def emptyline(self):
    # todo: improve this
    self.lines.append('\n')
    self.lines.append('\n')
  def newline(self):
    # todo: improve this
    if len(self.lines) > 0 and self.lines[-1] != '\n':
      self.lines.append('\n')
  def append(self, data):
    self.lines.append(data)
  def extend(self, data):
    self.lines.extend(data)
  def data(self, indent=None, first_indent=None, notrail=False):
    ret = ''.join(reduce(collapseLines, self.lines, []))
    if notrail and ret.endswith('\n'):
      ret = ret[:-1]
    if indent is None:
      return ret
    ret = newline_re.sub('\n' + indent.replace('\\', '\\\\') + '\\1', ret)
    if first_indent is False:
      return ret
    if first_indent is None:
      first_indent = indent
    return first_indent + ret

#------------------------------------------------------------------------------
class RstTranslator(nodes.GenericNodeVisitor):

  inline_format = {
    'inline'                          : ('{}',       None),
    'emphasis'                        : ('*{}*',     None),
    'strong'                          : ('**{}**',   None),
    'interpreted_or_phrase_ref'       : ('`{}`',     None),
    'title_reference'                 : ('`{}`',     None),
    'literal'                         : ('``{}``',   None),
    'inline_internal_target'          : ('_{}',      rstTicks),
    'footnote_reference'              : ('[{}]_',    None),
    'substitution_reference'          : ('|{}|',     None),
    'reference'                       : ('{}_',      rstTicks),
    'anonymous_reference'             : ('{}__',     rstTicks),
    }

  #----------------------------------------------------------------------------
  def __init__(self, document):
    nodes.NodeVisitor.__init__(self, document)
    self.settings = document.settings
    self.output   = Output()
    self.stack    = []
    self.level    = 0
    self.cache    = None
    self.cstack   = []

  #----------------------------------------------------------------------------
  def _pushStack(self):
    self.stack.append(self.output)
    self.output = Output()

  #----------------------------------------------------------------------------
  def _popStack(self):
    ret = self.output
    self.output = self.stack.pop()
    return ret

  #----------------------------------------------------------------------------
  def _putAttributes(self, node):
    title = None
    wtitle = False
    if isinstance(node, nodes.document):
      title = self.settings.title
      if title is not None:
        wtitle = True
    if not title:
      title = node.get('title')
    if title is not None:
      wtitle = self.settings.explicit_title or wtitle
      if not wtitle and isinstance(node, nodes.document):
        # check to see if the `document` title is different than
        # the inferred title, in which case we force-show the title
        # todo: is this really the best way to determine the inferred title?
        # todo: is checking for the first `title` node sufficient?
        # todo: there must be a way to use the rst parser code, no?...
        for subnode in node:
          if isinstance(subnode, nodes.title):
            if node.get('title') != subnode.astext():
              wtitle = True
            break
      if wtitle:
        self.output.emptyline()
        self.output.append('.. title:: ' + node['title'])
        self.output.emptyline()
    if node['classes']:
      self.output.emptyline()
      self.output.append('.. class:: ' + ' '.join(sorted(node['classes'])))
      self.output.emptyline()
    if node['ids']:
      # note: only generating an `id` node IFF they are not generated
      # todo: this is *not* the best way to determine whether or not
      #       the node['ids'] is completely generated!...
      dids = self.document.ids
      nids = node['ids']
      node['ids'] = []
      for id in nids:
        # todo: there is a weirdness that if the id references the document,
        #       then self.document.ids lookup points to a `section` node...
        #       look into what i misunderstood here.
        rnode = self.document.ids.get(id)
        if isinstance(node, nodes.document) and isinstance(rnode, nodes.section):
          rnode = node
        if rnode is node:
          del self.document.ids[id]
      self.document.set_id(node)
      nids_ng = set(nids) - set(node['ids'])
      if nids_ng:
        self.output.emptyline()
        for nid in sorted(nids_ng):
          self.output.append('.. _{id}:'.format(id=rstTicks(nid)))
          self.output.emptyline()
      self.document.ids = dids
      node['ids'] = nids

  #----------------------------------------------------------------------------
  def default_visit(self, node):
    if isinstance(node, nodes.Inline):
      self._pushStack()
    else:
      self._putAttributes(node)

  #----------------------------------------------------------------------------
  def default_departure(self, node):
    if isinstance(node, nodes.Inline):
      text = self._popStack().data()
      fmt = self.inline_format[node.__class__.__name__]
      if fmt[1]:
        text = fmt[1](text)
      self.output.append(fmt[0].format(text))

  #----------------------------------------------------------------------------
  def visit_problematic(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_problematic(self, node):
    self._popStack()
    text = rstTicks(node.astext())
    if text.startswith('`'):
      text = text[1:-1]
    # note: wrapping the link with newlines to protect
    # from other surrounding words.
    self.output.newline()
    self.output.append('`{text} <#{refuri}>`__'.format(
      text   = text,
      refuri = node['refid'],
      ))
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_system_message(self, node):
    kls = node['classes']
    node['classes'] = kls + ['system-message']
    self._putAttributes(node)
    node['classes'] = kls
    self.visit_title(None)
    self.output.append(
      '{type}/{level} ({source}, line {line})'.format(**node.attributes))
    self.depart_title(None)

  #----------------------------------------------------------------------------
  def depart_system_message(self, node):
    pass

  #----------------------------------------------------------------------------
  def depart_document(self, node):
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_Text(self, node):
    self.output.append(rstEscape(node.astext(), 'para'))

  #----------------------------------------------------------------------------
  def visit_title(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_title(self, node):
    sclen = len(self.settings.section_chars)
    over  = self.level < sclen
    lsym  = self.settings.section_chars[self.level % sclen]
    text  = self._popStack().data(notrail=True)
    if len(text) > 0 and text == text[0] * len(text) and re.match('[^a-zA-Z0-9]', text[0]):
      text = re.sub('([^a-zA-Z0-9])', '\\\\\\1', text)
    width = max(6, len(text))
    self.output.emptyline()
    if over:
      self.output.append(lsym * width)
      self.output.newline()
    self.output.append(text)
    self.output.newline()
    self.output.append(lsym * width)
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_section(self, node):
    self._putAttributes(node)
    self.level += 1

  #----------------------------------------------------------------------------
  def depart_section(self, node):
    self.level -= 1

  #----------------------------------------------------------------------------
  def visit_paragraph(self, node):
    self._putAttributes(node)
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_paragraph(self, node):
    text = self._popStack().data(notrail=True)
    self.output.emptyline()
    # todo: do textwrapping rules change in rST?...
    self.output.append(
      '\n'.join(textwrap.wrap(text, width=self.settings.text_width)))
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_literal_block(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_literal_block(self, node):
    text = self._popStack().data(notrail=True)
    self.output.emptyline()
    cmd = '::'
    if 'code' in node['classes']:
      cmd = '.. code-block::'
      if len(node['classes']) > 1:
        classes = node['classes'][:]
        classes.remove('code')
        cmd += ' ' + ' '.join(sorted(classes))
    self.output.append(cmd)
    self.output.emptyline()
    self.output.append(
      self.settings.indent + text.replace('\n', '\n' + self.settings.indent))
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_target(self, node):
    if isinstance(node.parent, nodes.TextElement) and node.referenced <= 1:
      # note: this is a little weird, but `target`s exist both when a link
      # is inlined AND when generated as
      return
    self.output.emptyline()
    self.output.append('.. _{name}: {uri}'.format(
      name = rstEscape(node['names'][0], context='`'),
      uri  = rstEscape(node['refuri'])))
    self.output.newline()

  #----------------------------------------------------------------------------
  def depart_target(self, node):
    pass

  #----------------------------------------------------------------------------
  def depart_reference(self, node):
    # todo: this is a bit of a hack... if this is an inline reference,
    #       i'm artificially inserting the "TEXT <URI>" output, and
    #       expecting the default handler to wrap it in "`...`_" --
    #       instead, there should be a helper method.
    sibs = list(node.parent)
    idx  = sibs.index(node)
    # todo: the ".lower()" is a little disconcerting here... is there
    #       a better way?...
    if idx + 1 < len(sibs) \
        and isinstance(sibs[idx + 1], nodes.target) \
        and node['name'].lower() in sibs[idx + 1]['names'] \
        and sibs[idx + 1].referenced == 1:
      text = self._popStack().data(notrail=True)
      self._pushStack()
      self.output.append('{text} <{uri}>'.format(
        text = text,
        uri  = rstEscape(node['refuri'])))
    else:
      if plaintexturi_re.match(node['refuri']):
        text = self._popStack().data()
        if node['refuri'] in (text, 'mailto:' + text):
          self.output.append(text)
          return
        # doh! something else! revert!...
        # todo: there *must* be a better explanation.
        self._pushStack()
        self.output.append(text)
    return self.default_departure(node)

  #----------------------------------------------------------------------------
  def visit_meta(self, node):
    sibs = list(node.parent)
    idx  = sibs.index(node)
    if idx == 0 or type(sibs[idx - 1]) is not type(node):
      self.output.emptyline()
      self.output.append('.. meta::')
      self.output.newline()
    self.output.append('{indent}:{name}: {content}'.format(
      indent  = self.settings.indent,
      name    = rstEscape(node['name'], context=':'),
      content = rstEscape(node['content']),
      ))
    self.output.newline()

  #----------------------------------------------------------------------------
  def depart_meta(self, node):
    pass

  #----------------------------------------------------------------------------
  def visit_bullet_list(self, node): pass
  def depart_bullet_list(self, node): pass

  #----------------------------------------------------------------------------
  def visit_list_item(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_list_item(self, node):
    blt  = node.parent.get('bullet', '*')
    text = self._popStack().data(
      indent=' ' * ( len(blt) + 1 ), first_indent=False, notrail=True)
    self.output.emptyline()
    self.output.append(blt + ' ' + rstEscape(text))
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_definition_list(self, node): pass
  def depart_definition_list(self, node): pass
  def visit_definition_list_item(self, node): pass
  def depart_definition_list_item(self, node): pass

  #----------------------------------------------------------------------------
  def visit_term(self, node):
    self.output.emptyline()

  #----------------------------------------------------------------------------
  def depart_term(self, node):
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_definition(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_definition(self, node):
    text = self._popStack().data(indent=self.settings.indent, notrail=True)
    self.output.emptyline()
    self.output.append(text)
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_comment(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_comment(self, node):
    text = self._popStack().data(
      indent=self.settings.indent, first_indent=False, notrail=True)
    self.output.emptyline()
    self.output.append('.. ' + rstEscape(text))
    self.output.newline()

  #----------------------------------------------------------------------------
  def _pushCache(self, cache):
    self.cstack.append(self.cache)
    self.cache = cache

  #----------------------------------------------------------------------------
  def _popCache(self):
    ret = self.cache
    self.cache = self.cstack.pop()
    return ret

  #----------------------------------------------------------------------------
  # TODO: this table rendering needs to be considerable improved!...
  #----------------------------------------------------------------------------

  #----------------------------------------------------------------------------
  def _getTableColspecs(self, node):
    while not isinstance(node, nodes.tgroup):
      node = node.parent
    return [n for n in node if isinstance(n, nodes.colspec)]

  #----------------------------------------------------------------------------
  def visit_table(self, node):
    self.output.emptyline()

  #----------------------------------------------------------------------------
  def depart_table(self, node):
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_thead(self, node):
    cspecs = self._getTableColspecs(node)
    self.output.newline()
    self.output.append(' '.join(['=' * c['colwidth'] for c in cspecs]))
    self.output.newline()

  #----------------------------------------------------------------------------
  def depart_thead(self, node):
    pass

  #----------------------------------------------------------------------------
  def visit_tbody(self, node):
    return self.visit_thead(node)

  #----------------------------------------------------------------------------
  def depart_tbody(self, node):
    return self.visit_thead(node)

  #----------------------------------------------------------------------------
  def visit_row(self, node):
    self._pushCache(list())

  #----------------------------------------------------------------------------
  def depart_row(self, node):
    entries = self._popCache()
    self.output.newline()
    cspecs = self._getTableColspecs(node)
    if len(cspecs) != len(entries):
      # TODO: tables can be more complex than this...
      # TODO: sound the alarm in a more docutils-ish way...
      raise ValueError('column count mismatch (colspecs: %r, row: %r)',
                       len(cspecs), len(entries))
    for idx, entry in enumerate(entries):
      spec = cspecs[idx]
      entry = entry.strip()
      if '\n' in entry:
        # TODO: tables can be more complex than this...
        # TODO: sound the alarm in a more docutils-ish way...
        raise ValueError('table cell entry contains a newline')
      if len(entry) > spec['colwidth']:
        # TODO: tables can be more complex than this...
        # TODO: sound the alarm in a more docutils-ish way...
        raise ValueError('table cell entry contents exceeds colwidth')
      self.output.append(entry)
      if idx + 1 < len(entries):
        self.output.append(' ' * ( spec['colwidth'] - len(entry) ))
        self.output.append(' ')
    self.output.newline()

  #----------------------------------------------------------------------------
  def visit_entry(self, node):
    self._pushStack()

  #----------------------------------------------------------------------------
  def depart_entry(self, node):
    text = self._popStack().data()
    self.cache.append(text)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/10/02
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, unittest
from . import rst

#------------------------------------------------------------------------------
class TestRst(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def rt(self, data, writer=None, settings=None):
    from docutils.core import publish_doctree, publish_from_doctree
    dt = publish_doctree(data, settings_overrides=settings)
    return publish_from_doctree(dt,
                                writer=writer or rst.Writer(),
                                settings_overrides=settings)

  #----------------------------------------------------------------------------
  def test_simple(self):
    src = '''\
.. class:: class1 class2

`Local` *Top-Level* ``10``
~~~~~~~~~~~~~~~~~~~~~~~~~~

some `link`_ foo.

Sub-Title
#########

more text

.. _`link`: http://foo.bar.com/
'''
    chk = '''\
.. class:: class1 class2

==========================
`Local` *Top-Level* ``10``
==========================

some link_ foo.

---------
Sub-Title
---------

more text

.. _link: http://foo.bar.com/
'''
    out = self.rt(src)
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_explicit_title(self):
    src = 'My *Title*\n----------\n\nsome text.\n'
    chk = '==========\nMy *Title*\n==========\n\nsome text.\n'
    out = self.rt(src)
    self.assertMultiLineEqual(out, chk)
    chk = '.. title:: My Title\n\n' + chk
    out = self.rt(src, settings={'explicit_title': True})
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_altered_title(self):
    from docutils import utils, nodes
    from docutils.core import publish_from_doctree
    doc = utils.new_document('<program>')
    doc['title']   = 'Altered Title'
    doc.append(nodes.title('', '', nodes.Text('Title')))
    doc.append(nodes.paragraph('', '', nodes.Text('some text.')))
    chk = '.. title:: Altered Title\n\n======\nTitle\n======\n\nsome text.\n'
    out = publish_from_doctree(
      doc, writer=rst.Writer(), settings_overrides={'explicit_title': False})
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_link(self):
    src = '''\
a `link with space`_.

.. _`link with space`: http://example.com/
'''
    self.assertMultiLineEqual(self.rt(src), src)

  #----------------------------------------------------------------------------
  def test_link_embedded(self):
    src = 'a `link with space <http://example.com>`_.'
    chk = 'a `link with space <http://example.com>`_.\n'
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_link_embedded_with_caps(self):
    src = 'a `Link With Space <http://example.com>`_.'
    chk = 'a `Link With Space <http://example.com>`_.\n'
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_literal_block(self):
    src = 'A literal example::\n\n  Code Line 1\n  ==> code line 2\n'
    chk = 'A literal example:\n\n::\n\n    Code Line 1\n    ==> code line 2\n'
    out = self.rt(src)
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_code_block(self):
    src = 'A literal example:\n\n.. code-block::\n\n  Code Line 1\n  ==> code line 2\n'
    chk = 'A literal example:\n\n.. code-block::\n\n    Code Line 1\n    ==> code line 2\n'
    out = self.rt(src)
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_code_block_python(self):
    try:
      import pygments
    except ImportError:
      sys.stderr.write('*** PYGMENTS LIBRARY NOT PRESENT - SKIPPING *** ')
      return
    src = 'A literal example:\n\n.. code-block:: python\n\n  import sys\n  sys.stdout.write("hello!")\n'
    chk = 'A literal example:\n\n.. code-block:: python\n\n    import sys\n    sys.stdout.write("hello!")\n'
    out = self.rt(src)
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_ids(self):
    from docutils import utils, nodes
    from docutils.core import publish_from_doctree
    doc = utils.new_document('<program>')
    doc['classes'] = ('c1 c2',)
    doc['ids'] = ('my-test-id',)
    doc.append(nodes.title('', '', nodes.Text('Title')))
    doc.append(nodes.paragraph('', '', nodes.Text('some text.')))
    doc.append(
      nodes.section(
        '',
        nodes.title('', '', nodes.Text('Sub-Title')),
        nodes.paragraph('', '', nodes.Text('some more text'))))
    chk = '''\
.. class:: c1 c2

.. _`my-test-id`:

======
Title
======

some text.

---------
Sub-Title
---------

some more text
'''
    out = publish_from_doctree(doc, writer=rst.Writer())
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_meta(self):
    from docutils import utils, nodes
    from docutils.parsers.rst.directives.html import MetaBody
    from docutils.core import publish_from_doctree
    doc = utils.new_document('<program>')
    doc.append(nodes.title('', '', nodes.Text('Title')))
    doc.append(nodes.paragraph('', '', nodes.Text('some text.')))
    doc.append(MetaBody('').meta('', name='title', content='Title'))
    doc.append(MetaBody('').meta('', name='generator', content='pyramid_describe/0.0.0'))
    doc.append(MetaBody('').meta('', name='location', content='http://example.com/'))
    chk = '''\
======
Title
======

some text.

.. meta::
    :title: Title
    :generator: pyramid_describe/0.0.0
    :location: http://example.com/
'''
    out = publish_from_doctree(
      doc, writer=rst.Writer(), settings_overrides={'explicit_title': False})
    self.assertMultiLineEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_lists(self):
    src = '''\
Current list of **states**
that this
engine can be in:

* beta
* production
* deprecated (this is a long line to cause
  the text handler to wrap it to the next line,
  and therefore test the indentation that must be
  done)

The following `skill` levels exist:

``Novice``:
  a true beginner.

``Intermediate``:
  an average user, *usually* expressed with::

    import average_user
    average_user.express()

``Expert``:
  the sky is the limit.
'''
    chk = '''\
Current list of **states** that this engine can be in:

* beta

* production

* deprecated (this is a long line to cause the text handler to wrap it to the
  next line, and therefore test the indentation that must be done)

The following `skill` levels exist:

``Novice``:

    a true beginner.

``Intermediate``:

    an average user, *usually* expressed with:

    ::

        import average_user
        average_user.express()

``Expert``:

    the sky is the limit.
'''
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_problematic(self):
    src = 'this paragraph is not **clean.\n'
    chk = '''\
this paragraph is not  `** <#id1>`__ clean.

.. class:: system-message

.. _id1:

============================
WARNING/2 (<string>, line 1)
============================

Inline strong start-string without end-string.
'''
    # TODO: it would be great to remove the error being printed out...
    #   <string>:1: (WARNING/2) Inline strong start-string without end-string.
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_paragraph_with_class(self):
    src = '.. class:: test beta\n\na paragraph.'
    chk = '.. class:: beta test\n\na paragraph.\n'
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_table(self):
    src = '''\
some text.

====================  =============================================
Cell 1                Column 2
====================  =============================================
``item 1.1``          item 1.2: **foo**
item 2.1              ``item 2.2``
====================  =============================================

some more text.
'''
    chk = '''\
some text.

==================== =============================================
Cell 1               Column 2
==================== =============================================
``item 1.1``         item 1.2: **foo**
item 2.1             ``item 2.2``
==================== =============================================

some more text.
'''
    self.assertMultiLineEqual(self.rt(src), chk)

## TODO: add support for grid tables...
#   #----------------------------------------------------------------------------
#   def test_table_grid(self):
#     src = '''\
# +------------+------------+-----------+
# | Header 1   | Header 2   | Header 3  |
# +============+============+===========+
# | body row 1 | column 2   | column 3  |
# +------------+------------+-----------+
# | body row 2 | Cells may span columns.|
# +------------+------------+-----------+
# | body row 3 | Cells may  | - Cells   |
# +------------+ span rows. | - contain |
# | body row 4 |            | - blocks. |
# +------------+------------+-----------+
# '''
#     chk = '''\
# +------------+------------+-----------+
# | Header 1   | Header 2   | Header 3  |
# +============+============+===========+
# | body row 1 | column 2   | column 3  |
# +------------+------------+-----------+
# | body row 2 | Cells may span columns.|
# +------------+------------+-----------+
# | body row 3 | Cells may  | - Cells   |
# +------------+ span rows. | - contain |
# | body row 4 |            | - blocks. |
# +------------+------------+-----------+
# '''
#     self.assertMultiLineEqual(self.rt(src), chk)

## TODO: add support for multi-row colspecs...
#   #----------------------------------------------------------------------------
#   def test_table_grid(self):
#     src = '''\
# =====  =====  ======
#    Inputs     Output
# ------------  ------
#   A      B    A or B
# =====  =====  ======
# False  False  False
# True   False  True
# False  True   True
# True   True   True
# =====  =====  ======
# '''
#     chk = '''\
# ===== ===== ======
# Inputs      Output
# ----------- ------
# A     B     A or B
# ===== ===== ======
# False False False
# True  False True
# False True  True
# True  True  True
# ===== ===== ======
# '''
#     self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_comment(self):
    src = 'some text.\n\n.. a comment: non-descript, eh?\n\nsome more text.'
    chk = 'some text.\n\n.. a comment: non-descript, eh?\n\nsome more text.\n'
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_nonlist_text(self):
    src = r'''
\(a) foo

\(1) foo

\(MCMLXX) foo

a\) foo

1\. foo

MCMLXX\) foo
'''
    chk = r'''\(a) foo

\(1) foo

\(MCMLXX) foo

\a) foo

\1. foo

\MCMLXX) foo
'''
    self.assertMultiLineEqual(self.rt(src), chk)

  #----------------------------------------------------------------------------
  def test_email(self):
    src = 'send your comments to asdf@example.com.\n'
    self.assertMultiLineEqual(self.rt(src), src)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
